/* easyrtc hook */
/*global module, require, async, _ */
module.exports = function (sails) {
    'use strict';
    // public attribute - holds application object
    this.appObj = null;
    var easyrtc = require('easyrtc');

    return {

        configure: function () {
            //console.log("--> Configure easyrtc Hook");

            // ICE Servers array
            var myIceConfig = [{url: "stun:stun.l.google.com:19302"},
                               {url: "stun:stun.sipgate.net"},
                               {url: "stun:217.10.68.152"},
                               {url: "stun:stun.sipgate.net:10000"},
                               {url: "stun:217.10.68.152:10000"},
                               {'url': 'stun:icall.zone:3478'}];
            // **Setting appIceConfig Across Server**
            easyrtc.setOption("appIceServers", myIceConfig);

			// helper function for room join event listener
			var updateOccupantsOnJoin = function (roomObj, connectionObj, roomName, roomStatus, done) {
				var appObj = connectionObj.getApp();
				var isRoomUser = false;

                if (!roomStatus.isReqAuth) { // this is random room from instant call
                    isRoomUser = true;
                } else {
                    // room is in DB and authentication is required, let's see options
                    if (!roomStatus.isAuthenticated && roomStatus.isHostInside) {
                        // room is not authenticated, but at least one room owner is in the room, so let the user enters but as GUEST
                        isRoomUser = false;
                    } else if (roomStatus.isAuthenticated) {
                        // room is authenticated, so this is the next Room User
                        isRoomUser = true;
                    } else {
                        // something went wrong
                        console.log(">>>>>>>>>>>>>>> OnRoomJoin event - room %s has no room owners!", roomName);
                        return done("You should not be able to reach this Room " + roomName);
                    }
                }

				async.parallel([
					function (next) {
						// setField --> isReqAuth
						roomObj.setField("isReqAuth", roomStatus.isReqAuth, {isShared: true}, function (err) {
                            if (err) {
                                return next(err);
                            }
							return next(null);
						});
					},
					function (next) {
						async.waterfall([
							function (next) {
								// getField --> roomOccupants
								roomObj.getField("roomOccupants", function (err, roomOccupants) {
                                    if (err) {
                                        return next(err);
                                    }
                                    connectionObj.util.logDebug("[" + appObj.getAppName() + "]" +
																(connectionObj ? "[" + connectionObj.getEasyrtcid() + "]" : "") +
																" Running func 'getField' for 'roomOccupants' field " + JSON.stringify(roomOccupants.fieldValue));

                                    var roomOccupantsObj = roomOccupants.fieldValue;

									return next(null, roomOccupantsObj);
								});
							},
							function (roomOccupantsObj, next) {
								// update users & guests
								if (!isRoomUser && !roomOccupantsObj.users.length) {
                                    // the current connection / user is not Room User and no available Room User, so throw error
                                    console.log(">>>>>>>>>>>>>>> OnRoomJoin event - room %s has no room owners!", roomName);
                                    return next("Room " + roomName + " has no room owners!");
                                }

                                // check to see if authorization was successful --> add connection to Room Users array
                                if (isRoomUser) {
                                    // just add this connection easyrtcid as last element to users array
                                    roomOccupantsObj.users[roomOccupantsObj.users.length] = connectionObj.getEasyrtcid();
                                } else {
                                    // otherwise add connection easyrtcid to guests array
                                    roomOccupantsObj.guests[roomOccupantsObj.guests.length] = connectionObj.getEasyrtcid();
                                }
								return next(null, roomOccupantsObj);
							},
							function (roomOccupantsObj, next) {
								// setField --> roomOccupants
                                connectionObj.util.logDebug("[" + appObj.getAppName() + "]" + (connectionObj ? "[" +
															connectionObj.getEasyrtcid() + "]" : "") +
															" Running func 'setField' for 'roomOccupants' field " + JSON.stringify(roomOccupantsObj));

								roomObj.setField("roomOccupants", roomOccupantsObj, {isShared: true}, function (err) {
                                    if (err) {
                                        return next(err);
                                    }
									return next(null);
								});
							},
							function (next) {
								// emit data
								roomObj.emitRoomDataFieldUpdate(connectionObj.getEasyrtcid(), function (err) {
                                    if (err) {
                                        return next(err);
                                    }

                                    connectionObj.util.logDebug("[" + appObj.getAppName() + "]" +
																(connectionObj ? "[" + connectionObj.getEasyrtcid() + "]" : "") +
																" Field [roomOccupants] update sent to all in room " + roomName);
									return next(null);
								});
							}
						], function (err, results) {
							if (err) {
								return next(err);
							}
							return next(null);
						});
					}
				], function (err, results) {
					if (err) {
						return done(err);
					}

					return done(null);
				});
			};

			// helper function for room join event listener
			var updateSubscribersOnJoin = function (roomObj, connectionObj, roomName, roomParameter, done) {
				var appObj = connectionObj.getApp();

				async.waterfall([
					function (next) {
						// getField --> roomSubscribers
						roomObj.getField("roomSubscribers", function (err, roomSubscribers) {
                            if (err) {
                                return next(err);
                            }

                            connectionObj.util.logDebug("[" + appObj.getAppName() + "]" + (connectionObj ? "[" +
														connectionObj.getEasyrtcid() + "]" : "") +
														" Running func 'getField' for 'roomSubscribers' field " + JSON.stringify(roomSubscribers.fieldValue));
							var roomSubscribersObj = roomSubscribers.fieldValue;

							return next(null, roomSubscribersObj);
						});
					},
					function (roomSubscribersObj, next) {
						// update publishers & subscribers
						if (roomParameter.pubsub === 'publisher') {
                            roomSubscribersObj.publishers[roomSubscribersObj.publishers.length] = connectionObj.getEasyrtcid();
                        } else if (roomParameter.pubsub === 'subscriber') {
                            roomSubscribersObj.subscribers[roomSubscribersObj.subscribers.length] = connectionObj.getEasyrtcid();
                        } else {
                            return next('Not valid pubsub value!!!');
                        }

						return next(null, roomSubscribersObj);
					},
					function (roomSubscribersObj, next) {
						// setField --> roomSubscribers
						connectionObj.util.logDebug("[" + appObj.getAppName() + "]" + (connectionObj ? "[" +
													connectionObj.getEasyrtcid() + "]" : "") +
													" Running func 'setField' for 'roomSubscribers' field " + JSON.stringify(roomSubscribersObj));

                        roomObj.setField("roomSubscribers", roomSubscribersObj, {isShared: true}, function (err) {
                            if (err) {
                                return next(err);
                            }
							return next(null);
						});
					},
					function (next) {
						// emit data
						roomObj.emitRoomDataFieldUpdate(connectionObj.getEasyrtcid(), function (err) {
                            if (err) {
                                return next(err);
                            }

                            connectionObj.util.logDebug("[" + appObj.getAppName() + "]" + (connectionObj ? "[" +
														connectionObj.getEasyrtcid() + "]" : "") +
														" Field [roomSubscribers] update sent to all in room " + roomName);
							return next(null);
						});
					}
				], function (err, results) {
					if (err) {
						return done(err);
					}
					return done(null);
				});
			};

			// custom listener for onRoomJoin event
			var onRoomJoin = function (connectionObj, roomName, roomParameter, callback) {
				connectionObj.util.logDebug("[" + connectionObj.getAppName() +
											"][" + connectionObj.getEasyrtcid() +
											"] Running func 'onRoomJoin' for room [" +
											roomName + "] Room params [" + JSON.stringify(roomParameter) + "]");

				if (!roomParameter.pubsub && !roomParameter.status) {
					// something went wrong - return error and not allow room join
                    return callback('joinRoom called without proper room parameters');
				}

				var appObj = connectionObj.getApp();
				async.series([
					function (done) {
						// join room
						connectionObj.joinRoom(roomName, function(err, connectionRoomObj){
							if (err) {
								return done(err);
							}

							return done(null, connectionRoomObj);
						});
					},
					function (done) {
						// set user name
						connectionObj.setUsername(roomParameter.user, function (err) {
							if (err) {
								return done(err);
							}
							return done(null);
						});
					},
					function (done) {
						async.waterfall([
							function (done) {
								// get room object
								appObj.room(roomName, function (err, roomObj) {
									connectionObj.util.logDebug("[" + appObj.getAppName() + "]" +
																(connectionObj ? "[" + connectionObj.getEasyrtcid() + "]" : "") +
																" Retrieve property 'room'");
									if (err) {
										return done(err);
									}

									return done(null, roomObj);
								});
							},
							function (roomObj, done) {
								// update room occupants & subscribers
								async.parallel([
									function (done) {
										// update roomOccupants
										if (roomParameter.status) {
											return updateOccupantsOnJoin(roomObj, connectionObj, roomName, roomParameter.status, done);
										} else {
											return done(null);
										}
									},
									function (done) {
										// update roomSubscribers
										if (roomParameter.pubsub) {
											return updateSubscribersOnJoin(roomObj, connectionObj, roomName, roomParameter, done);
										} else {
											return done(null);
										}
									}
								], function (err, results) {
									if (err) {
										return done(err);
									}
									return done(null, results);
								});
							}
						], function (err, results) {
							if (err) {
								return done(err);
							}

							return done(null, results);
						});
					}
				],
				function (err, results) {
					if (err) {
						return callback(err);
					}

					var connectionRoomObj = results[0];

					connectionRoomObj.emitRoomDataDelta(false, function (err, roomDataDelta) {
						// Return connectionRoomObj regardless of if there was a problem sending out the deltas
						return callback(null, connectionRoomObj);
					});
				});
			};

            easyrtc.events.on("roomJoin", onRoomJoin);

            var updateOccupantsOnLeave = function(roomObj, connectionObj, done) {
                var appObj = connectionObj.getApp();
                async.waterfall([
                    function (next) {
                        // get field
                        roomObj.getField("roomOccupants", function (err, roomOccupants) {
                            if (err) {
                                return next(err);
                            }
                            connectionObj.util.logDebug("[" + appObj.getAppName() +
                                                        "]" + (connectionObj ? "[" +
                                                               connectionObj.getEasyrtcid() + "]" : "") +
                                                        " Running func 'getField' for 'roomOccupants' field " +
                                                        JSON.stringify(roomOccupants.fieldValue));

                            // revert stringified object to real object
                            var roomOccupantsObj = roomOccupants.fieldValue;

                            // check to see what is this connection easyrtcid
                            var index = roomOccupantsObj.users.indexOf(connectionObj.getEasyrtcid());
                            // check if connection is Call Owner
                            if (!index) {
                                // found call owner

                                // ok, just shift the rest of users' array, so the next Room User will become Call Owner
                                roomOccupantsObj.users.shift();

                            } else if(index > 0) {
                                // room user leaving the room, so just remove it from the users array
                                roomOccupantsObj.users.splice(index, 1);
                            }
                            else if(index === -1) {
                                // connection is not inside users array, so it is inside guests array, check it
                                index = roomOccupantsObj.guests.indexOf(connectionObj.getEasyrtcid());
                                if(index >= 0) {
                                    // remove guest from guests array
                                    roomOccupantsObj.guests.splice(index, 1);
                                }
                            }

                            return next(null, roomOccupantsObj);
                        });
                    },
                    function (roomOccupantsObj, next) {
                        // set field
                        connectionObj.util.logDebug("[" +
                                                    appObj.getAppName() + "]" +
                                                    (connectionObj?"["+connectionObj.getEasyrtcid()+"]":"") +
                                                    " Running func 'setField' for 'roomOccupants' field " +
                                                    JSON.stringify(roomOccupantsObj));
                        roomObj.setField("roomOccupants", roomOccupantsObj, {isShared: true}, function(err) {
                            if (err) {
                                return next(err);
                            }

                            return next(null, roomOccupantsObj.users.length + roomOccupantsObj.guests.length);
                        });
                    },
                    function (roomOccupants, next) {
                        // emit data
                        roomObj.emitRoomDataFieldUpdate(/*connectionObj.getEasyrtcid()*/ null, function(err) {
                            if (err) {
                                return next(err);
                            }

                            connectionObj.util.logDebug("[" +
                                                        appObj.getAppName() + "]" +
                                                        (connectionObj?"[" + connectionObj.getEasyrtcid() +
                                                         "]":"") + " Field update sent to all in room " +
                                                        roomObj.getRoomName());

                            // all went fine, just return roomOccupantsObj as resutls
                            return next(null, roomOccupants);
                        });
                    }
                ], function (err, results) {
                    // return occupantCount
                    if (err) {
                        return done(err);
                    }
                    return done(null, results);
                });
            };

            var updateSubscribersOnLeave = function(roomObj, connectionObj, done) {
                var appObj = connectionObj.getApp();
                async.waterfall([
                    function (next) {
                        // get field
                        roomObj.getField("roomSubscribers", function (err, roomSubscribers) {
                            if (err) {
                                return next(err);
                            }
                            connectionObj.util.logDebug("[" + appObj.getAppName() +
                                                        "]" + (connectionObj ? "[" +
                                                        connectionObj.getEasyrtcid() + "]" : "") +
                                                        " Running func 'getField' for 'roomSubscribers' field " +
                                                        JSON.stringify(roomSubscribers.fieldValue));

                            // revert stringified object to real object
                            var roomSubscribersObj = roomSubscribers.fieldValue;

                            // check to see what is this connection easyrtcid
                            var index = roomSubscribersObj.publishers.indexOf(connectionObj.getEasyrtcid());
                            // check if connection is Call Owner
                            if(index >= 0) {
                                // room publisher leaving the room, so just remove it from the publishers array
                                roomSubscribersObj.publishers.splice(index, 1);
                            } else if(index == -1) {
                                // connection is not inside publishers array, so it is inside subscribers array, check it
                                index = roomSubscribersObj.subscribers.indexOf(connectionObj.getEasyrtcid());
                                if(index >= 0) {
                                    // remove guest from guests array
                                    roomSubscribersObj.subscribers.splice(index, 1);
                                }
                            }

                            return next(null, roomSubscribersObj);
                        });
                    },
                    function (roomSubscribersObj, next) {
                        // set field
                        connectionObj.util.logDebug("[" +
                                                    appObj.getAppName() + "]" +
                                                    (connectionObj?"["+connectionObj.getEasyrtcid()+"]":"") +
                                                    " Running func 'setField' for 'roomSubscribers' field " +
                                                    JSON.stringify(roomSubscribersObj));
                        roomObj.setField("roomSubscribers", roomSubscribersObj, {isShared: true}, function(err) {
                            if (err) {
                                return next(err);
                            }

                            return next(null, roomSubscribersObj.publishers.length +
                                        roomSubscribersObj.subscribers.length);
                        });
                    },
                    function (roomSubscribers, next) {
                        // emit data
                        roomObj.emitRoomDataFieldUpdate(/*connectionObj.getEasyrtcid()*/ null, function(err) {
                            if (err) {
                                return next(err);
                            }

                            connectionObj.util.logDebug("[" +
                                                        appObj.getAppName() + "]" +
                                                        (connectionObj?"[" + connectionObj.getEasyrtcid() +
                                                         "]":"") + " Field update sent to all in room " +
                                                        roomObj.getRoomName());

                            // all went fine, just return roomSubscribers as resutls
                            return next(null, roomSubscribers);
                        });
                    }
                ], function (err, results) {
                    // return subscribersCount
                    if (err) {
                        return done(err);
                    }
                    return done(null, results);
                });
            };

            var onRoomLeave = function(connectionObj, roomName, next){
                if(!_.isFunction(next)) {
                    next = connectionObj.util.nextToNowhere;
                }
                var appObj = connectionObj.getApp();
                async.waterfall([
                    function (done) {
                        // get roomObj
                        appObj.room(roomName, function (err, roomObj) {
                            connectionObj.util.logDebug("[" + appObj.getAppName() + "]" +
                                                        (connectionObj ? "[" + connectionObj.getEasyrtcid() + "]" : "") +
                                                        " Retrieve roomObj");
                            if (err) {
                                return done(err);
                            }
                            return done(null, roomObj);
                        });
                    },
                    function (roomObj, done) {
                        async.parallel([
                            function (done) {
                                // get roomOccupants
                                return updateOccupantsOnLeave(roomObj, connectionObj, done);
                            },
                            function (done) {
                                // get roomSubscribers
                                return updateSubscribersOnLeave(roomObj, connectionObj, done);
                            }
                        ], function (err, results) {
                            if (err) {
                                return done(err);
                            }

                            return done(null, results[0], results[1]);
                        });
                    },
                    function (roomOccupants, roomSubscribers, done) {
                        // get connectionRoomObj --> leave room
                        connectionObj.util.logDebug("[" + connectionObj.getAppName()
                                                    + "][" + connectionObj.getEasyrtcid()
                                                    + "] Running func 'onRoomLeave' with rooms ["
                                                    + roomName + "]");

                        connectionObj.room(roomName, function(err, connectionRoomObj){
                            if (err) {
                                connectionObj.util.logWarning("[" + connectionObj.getAppName()
                                                              + "][" + connectionObj.getEasyrtcid()
                                                              + "] Couldn't leave room [" + roomName + "]");
                                return done(err);
                            }

                            connectionObj.util.logDebug("[" + connectionObj.getAppName()
                                                        + "][" + connectionObj.getEasyrtcid()
                                                        + "] Leave room [" + roomName + "]");
                            connectionRoomObj.leaveRoom(function (err) {
                                if (err) {
                                    return done(err);
                                }

                                return done(null, roomOccupants, roomSubscribers);
                            });
                        });
                    },
                    function (roomOccupants, roomSubscribers, done) {
                        // check to see if room still have some occupants
                        if (!roomOccupants) {
                            // if all room occupants left, delete the room
                            appObj.deleteRoom(roomName, function (err, isDeleted) {
                                if (err) {
                                    return done(err);
                                }
                                connectionObj.util.logDebug("[" + connectionObj.getAppName() + "][" +
                                                            connectionObj.getEasyrtcid() + "] Delete room [" +
                                                            roomName + "] = " + isDeleted);
                                return done(null, roomOccupants, roomSubscribers);
                            });
                        } else {
                            return done(null, roomOccupants, roomSubscribers);
                        }
                    }
                ], function (err, results) {
                    if (err) {
                        return next(err);
                    }
                    return next(null);
                });
            };

            easyrtc.events.on("roomLeave", onRoomLeave);

			// custom listener for disconnect
			var onDisconnect = function(connectionObj, next){
				connectionObj.util.logDebug("Running custom func 'onDisconnect'");

				async.waterfall([
					function(asyncCallback) {
						// Get array of rooms
						connectionObj.getRoomNames(asyncCallback);
					},
					function(roomNames, asyncCallback) {
						// leave all rooms
						async.each(roomNames,
							function(currentRoomName, asyncEachCallback) {
								connectionObj.events.emit("roomLeave", connectionObj, currentRoomName, function(err){asyncEachCallback(null);});
							},
							function(err){
								asyncCallback(null);
							}
						);
					},
					// notify all other connections the current is about to leave
					function (asyncCallback) {
						EasyrtcService.notifyOnRoomLeave({leavingConnection: connectionObj.getEasyrtcid()}, asyncCallback);
					},
					function(asyncCallback) {
						// log all connections as ended
						connectionObj.util.logDebug("["+connectionObj.getAppName()+"]["+connectionObj.getEasyrtcid()+"] Disconnected Custom");
						connectionObj.removeConnection(asyncCallback);
					}
				], function(err) {
					next(null);
				});

				next(null);
			};
            easyrtc.events.on("disconnect", onDisconnect);
        },
        initialize: function (cb) {
            var eventsToWaitFor = ['hook:sockets:loaded', 'hook:http:loaded'];

            function listenCallback(err, pub) {
                pub.util.logDebug("--> Running func 'listenCallback' ");

                pub.createApp("easysails", {"roomDefaultEnable": false}, function (err, appObj) {
                    if (err) { return; }

                    sails.hooks.easyrtc.appObj = appObj;

                    //console.log('>>> easyrtc hook initialize: hook.appObj is ' + sails.hooks.easyrtc.appObj);
                });

                pub.setOption("roomDefaultFieldObj",
                              {"isReqAuth": {fieldValue: false, fieldOption: {isShared: true}},
                               "roomOccupants": {fieldValue: {users: [], guests: []}, fieldOption: {isShared: true}},
                               "roomSubscribers": {fieldValue: {publishers: [], subscribers: []},
                                                   fieldOption: {isShared: true}}
                              });
            }

            // create own socket server
            //var socketServer = sails.io.listen(sails.hooks.http.server, {"log level":1});

            sails.after(eventsToWaitFor, function () {
                // Start EasyRTC server with options to change the log level and add dates to the log.
                var logLevel = sails.config.environment === 'production' ? 'info' : 'debug',
                    easyrtcServer = easyrtc.listen(sails.hooks.http.app,
                                                   sails.io,
                                                   {logLevel: logLevel, logDateEnable: true},
                                                   listenCallback);
				easyrtc.setOption("demosEnable", sails.config.environment === 'development');

                // Finish initializing custom hook
                // Then call cb()
                return cb();
            });
        }
    };
};
